from abc import ABC, abstractmethod
from typing import Any, Callable, Generic, Iterable, Type, TypeVar

from .enums import Scope
from .parameter_parser import get_parameters
from .parameters import Parameter

_T = TypeVar("_T")
Constructor = Callable[..., _T]


class PieceData(ABC, Generic[_T]):
    __slots__ = ("type", "_constructor", "parameters", "_instance")

    def __init__(self, type: Type[_T], constructor: Constructor[_T]) -> None:
        self.type: Type[_T] = type
        self._constructor = constructor
        self.parameters: Iterable[Parameter] = get_parameters(constructor)
        self._instance: _T | None = None

    @abstractmethod
    def get_instance(self) -> _T | None:
        """Simple getter for the instance of the piece.

        Returns
        -------
        _T | None
            instance if it exists, otherwise None
        """

    @abstractmethod
    def initialize(self, parameters: dict[str, Any]) -> _T:
        """Initializes instance from specified PieceData with help of self._constructor.

        Parameters
        ----------
        parameters : dict[str, Any]
            _description_

        Returns
        -------
        _T
            created instance
        """


class OriginalPieceData(PieceData[_T]):
    def get_instance(self) -> _T | None:
        return None

    def initialize(self, parameters) -> _T:
        return self._constructor(**parameters)


class UniversalPieceData(PieceData[_T]):
    def get_instance(self) -> _T | None:
        return self._instance

    def initialize(self, parameters) -> _T:
        self._instance = self._constructor(**parameters)
        return self._instance


piece_data_mapping = {
    Scope.UNIVERSAL: UniversalPieceData,
    Scope.ORIGINAL: OriginalPieceData,
}


def piece_data_factory(
    type_: Type[_T], scope: Scope, constructor: Constructor
) -> PieceData[_T]:
    return piece_data_mapping[scope][_T](type_, constructor)



from .enums import InitStrategy, Scope
from .exceptions import (
    AmbiguousPieceException,
    ParameterNotAnnotatedException,
    PieceException,
    PieceIncorrectUseException,
    PieceNotFound,
)
from .facade import (
    Piece,
    PieceFactory,
    get_piece,
    get_pieces_by_name,
    get_pieces_by_supertype,
    register_piece,
    register_piece_factory,
)

__all__ = [
    "Piece",
    "PieceFactory",
    "get_piece",
    "register_piece",
    "register_piece_factory",
    "get_pieces_by_name",
    "get_pieces_by_supertype",
    "PieceException",
    "PieceNotFound",
    "ParameterNotAnnotatedException",
    "AmbiguousPieceException",
    "PieceIncorrectUseException",
    "InitStrategy",
    "Scope",
]



import re
from inspect import _empty, signature
from typing import Any, Callable, Iterator, ParamSpec, Type, TypeVar

from .core import piece_data_factory
from .enums import InitStrategy
from .enums import Scope
from .exceptions import PieceIncorrectUseException
from .registry import registry

_T = TypeVar("_T")
P = ParamSpec("P")

LAZY = InitStrategy.LAZY


def _track_piece(
    piece_type: Type[_T],
    piece_name: str,
    constructor: Callable[..., _T],
    creation_type: InitStrategy = LAZY,
    scope: Scope = Scope.UNIVERSAL,
) -> None:
    if (scope, creation_type) == (Scope.ORIGINAL, InitStrategy.EAGER):
        raise PieceIncorrectUseException(
            "ORIGINAL scope with EAGER creation strategy is illegal"
        )

    piece_data = piece_data_factory(piece_type, scope, constructor)

    registry.add(piece_name, piece_data)

    if creation_type == InitStrategy.EAGER:
        registry.get_object(piece_name, piece_type)


def get_piece(piece_name: str, piece_type: Type[_T]) -> _T:
    """This function returns registered piece by name and type.

    Parameters
    ----------
    piece_name : str
        name of the piece
    piece_type : Type[_T]

    Returns
    -------
    T
        desired instance
    """
    return registry.get_object(piece_name, piece_type)


def get_pieces_by_supertype(super_type: Type[_T]) -> Iterator[_T]:
    """This function returns all registered pieces that are subtypes of given type.

    Parameters
    ----------
    super_type : Type[T]

    Returns
    -------
    T
        desired instance
    """
    return registry.get_all_objects_by_supertype(super_type)


def get_pieces_by_name(name_pattern: str | re.Pattern[str]) -> Iterator[Any]:
    """This function returns all registered pieces that match given name pattern.

    Parameters
    ----------
    name_pattern : str | re.Pattern[str]
        regular expression pattern to match name

    Returns
    -------
    Iterator[Any]
        iterator of instances matching name
    """
    return registry.get_all_objects_by_name_matching(re.compile(name_pattern))


def register_piece(
    cls: Type[_T],
    piece_name: str,
    creation_type: InitStrategy = InitStrategy.LAZY,
    scope: Scope = Scope.UNIVERSAL,
) -> None:
    """This function registers class as a dependency.
    __init__ method's parameters must be annotated references to registered pieces.

    Parameters
    ----------
    name : str
        unique name of piece
    creation_type : InitStrategy, optional
        defines instantiation strategy, by default `InitStrategy.LAZY`\\
        `LAZY` - object is created on first access\\
        `EAGER` - object is created immediately
    scope : Scope, optional
        scope of the piece, by default Scope.UNIVERSAL\\
        `ORIGINAL` - piece is created only once and is shared among all usages\\
        `UNIVERSAL` - piece is created for each usage separately
    """
    _track_piece(cls, piece_name, cls, creation_type, scope)


def register_piece_factory(
    factory: Callable[..., _T],
    name: str | None,
    creation_type: InitStrategy = LAZY,
    scope: Scope = Scope.UNIVERSAL,
) -> None:
    """This function registers a factory function to create dependency.\\
    Factory function's parameters must be annotated references to registered pieces.\\
    Factory function must declare return type.\\
    **Tip**: Use `PieceFactory` decorator instead.

    Parameters
    ----------
    factory : Callable[..., T]
        factory function to be registered
    name : str | None, optional
        name of piece, if None piece piece has same name as factory function, by default None
    creation_type : InitStrategy, optional
        defines instantiation strategy, by default `InitStrategy.LAZY`\\
        `LAZY` - object is created on first access\\
        `EAGER` - object is created immediately
    scope : Scope, optional
        scope of the piece, by default Scope.UNIVERSAL\\
        `ORIGINAL` - piece is created only once and is shared among all usages\\
        `UNIVERSAL` - piece is created for each usage separately
    """
    piece_type = signature(factory).return_annotation

    if piece_type is _empty or piece_type is None:
        raise PieceIncorrectUseException(
            f"Function `{factory.__name__}` must have return type specified and cannot be None"
        )

    _track_piece(
        piece_type,
        name if name is not None else factory.__name__,
        factory,
        creation_type,
        scope,
    )


def Piece(
    name: str,
    init_strategy: InitStrategy = LAZY,
    scope: Scope = Scope.UNIVERSAL,
):
    """This decorator registers class as a dependency.
    __init__ method's parameters must be annotated references to registered pieces.

    Parameters
    ----------
    name : str
        unique name of piece
    creation_type : InitStrategy, optional
        defines instantiation strategy, by default `InitStrategy.LAZY`\\
        `LAZY` - object is created on first access\\
        `EAGER` - object is created immediately
    scope : Scope, optional
        scope of the piece, by default Scope.UNIVERSAL\\
        `ORIGINAL` - piece is created only once and is shared among all usages\\
        `UNIVERSAL` - piece is created for each usage separately
    """

    def inner(cls: Type[_T]) -> Type[_T]:
        register_piece(cls, name, init_strategy, scope)
        return cls

    return inner


def PieceFactory(
    name: str | None = None,
    init_strategy: InitStrategy = InitStrategy.LAZY,
    scope: Scope = Scope.UNIVERSAL,
):
    """This decorator registers a factory function to create dependency.\\
    Factory function's parameters must be annotated references to registered pieces.\\
    Factory function must declare return type.

    Parameters
    ----------
    name : str | None, optional
        name of piece, if None piece piece has same name as factory function, by default None
    creation_type : InitStrategy, optional
        defines instantiation strategy, by default `InitStrategy.LAZY`\\
        `LAZY` - object is created on first access\\
        `EAGER` - object is created immediately
    scope : Scope, optional
        scope of the piece, by default Scope.UNIVERSAL\\
        `ORIGINAL` - piece is created only once and is shared among all usages\\
        `UNIVERSAL` - piece is created for each usage separately
    """

    def inner(factory: Callable[P, _T]) -> Callable[P, _T]:
        register_piece_factory(factory, name, init_strategy, scope)
        return factory

    return inner



import inspect
from typing import Annotated, Any, Callable, ForwardRef, Iterable

from .exceptions import (
    ParameterNotAnnotatedException,
    PieceException,
    PieceIncorrectUseException,
)
from .parameters import (
    DefaultFactoryParameter,
    DefaultParameter,
    Parameter,
    PieceParameter,
)

ANNOTATION_TYPE = type(Annotated[str, "example"])


def create_piece_parameter(
    name: str, piece_type: Any, piece_name: str
) -> PieceParameter:
    if not piece_name.strip():
        raise PieceException("piece_name must not be blank")
    return PieceParameter(name, piece_name, piece_type)


def create_default_factory_parameter(name: str, factory: Callable[[], Any]):
    return DefaultFactoryParameter(name, factory)


def evaluate_forward_ref(fr: ForwardRef, globals_dict: dict[str, Any]) -> Any:
    raise PieceException("ForwardRef is not supported")


def count_non_default_parameters(fn) -> int:
    try:
        parameters = inspect.signature(fn).parameters.values()
    except ValueError:
        parameters = tuple()

    filtered = filter(
        lambda p: p.default is inspect.Parameter.empty,
        parameters,
    )
    return sum(1 for _ in filtered)


def parse_parameter(parameter: inspect.Parameter) -> Parameter:
    annotation = parameter.annotation

    if parameter.default is not inspect.Parameter.empty:
        return DefaultParameter(parameter.name, parameter.default)

    if type(annotation) is not ANNOTATION_TYPE:
        raise ParameterNotAnnotatedException(parameter)

    metadata = annotation.__metadata__
    if len(metadata) < 1:
        raise PieceIncorrectUseException("piece metadata not specified in Annotated[]")

    piece_type = annotation.__origin__

    if isinstance(piece_type, ForwardRef):
        try:
            gd = metadata[1]
            assert isinstance(gd, dict), "expected globals to be instance of dict"
        except IndexError:
            raise PieceException("globals not provided to evaluate ForwardRef")
        except AssertionError as e:
            raise PieceException(e.args[0])

        piece_type = evaluate_forward_ref(piece_type, gd)

    metainfo = metadata[0]
    if isinstance(metainfo, str):
        return create_piece_parameter(parameter.name, piece_type, metainfo)

    if callable(metainfo) and count_non_default_parameters(metainfo) == 0:
        return create_default_factory_parameter(parameter.name, metainfo)

    raise PieceIncorrectUseException("invalid use")


def get_parameters(fn: Callable[..., Any]) -> Iterable[Parameter]:
    return tuple(map(parse_parameter, inspect.signature(fn).parameters.values()))



from inspect import Parameter
from typing import Any, Type


class PieceException(Exception):
    def __init__(self, message: str) -> None:
        self.message = message

    def __str__(self) -> str:
        return self.message


class PieceIncorrectUseException(PieceException):
    pass


class PieceNotFound(PieceException):
    pass


class ParameterNotAnnotatedException(PieceException):
    def __init__(self, parameter: Parameter) -> None:
        super().__init__(
            f"Parameter `{parameter.name}` is not annotated with `typing.Annotated` (actual type: {parameter.annotation})"
        )


class AmbiguousPieceException(PieceException):
    pass


class _NeedCalculation(RuntimeError):
    def __init__(self, piece_name: str, piece_type: Type[Any]) -> None:
        self.piece_name = piece_name
        self.piece_type = piece_type



from collections import defaultdict
from re import Pattern
from typing import Any, Iterator, Type, TypeVar

from .core import PieceData
from .exceptions import (
    AmbiguousPieceException,
    PieceNotFound,
    _NeedCalculation,
)

Storage = dict[str, dict[Type[Any], PieceData[Any]]]

_T = TypeVar("_T")


class Registry:
    def __init__(self):
        self.registry: Storage = defaultdict(dict)

    def add(self, piece_name: str, piece_data: PieceData[Any]):
        piece_dict = self.registry[piece_name]

        if piece_data.type in piece_dict:
            raise AmbiguousPieceException(
                f"Piece {piece_data.type} is already registered as a subclass of {piece_data.type}."
            )

        piece_dict[piece_data.type] = piece_data

    def _get_piece_data(
        self, piece_name: str, piece_type: Type[_T]
    ) -> PieceData[_T] | None:
        if (pd := self.registry[piece_name].get(piece_type)) is not None:
            return pd

        for type_, pd in self.registry[piece_name].items():
            if issubclass(type_, piece_type):
                return pd
        return None

    def get_object(self, piece_name: str, piece_type: Type[_T]) -> _T:
        piece_data = self._get_piece_data(piece_name, piece_type)

        if piece_data is None:
            raise PieceNotFound(f"Piece {piece_type} not found in registry.")

        if (instance := piece_data.get_instance()) is not None:
            return instance

        params: dict[str, Any] = {}
        for param in piece_data.parameters:
            try:
                param_val = param.get()
            except _NeedCalculation as e:
                param_val = self.get_object(e.piece_name, e.piece_type)
            params[param.name] = param_val

        return piece_data.initialize(params)

    def get_all_objects_by_supertype(self, super_type: Type[_T]) -> Iterator[_T]:
        for piece_name, piece_data in self.registry.items():
            for type_ in piece_data.keys():
                if issubclass(type_, super_type):
                    yield self.get_object(piece_name, type_)

    def get_all_objects_by_name_matching(self, name_pattern: Pattern) -> Iterator[Any]:
        for name, data in (
            (name, data)
            for name, data in self.registry.items()
            if name_pattern.search(name)
        ):
            for piece_data in data.values():
                yield self.get_object(name, piece_data.type)

    def clear(self):
        self.registry.clear()

    def __getitem__(self, item: str) -> dict[Type[Any], PieceData[Any]]:
        return self.registry[item]


registry: Registry = Registry()



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Callable, Type

from .exceptions import _NeedCalculation


@dataclass(frozen=True)
class AbstractFrozenDataclass(ABC):
    def __new__(cls, *args, **kwargs):
        if (
            cls is AbstractFrozenDataclass
            or cls.__bases__[0] == AbstractFrozenDataclass
        ):
            raise TypeError("Cannot instantiate abstract class.")
        return super().__new__(cls)


@dataclass(frozen=True, slots=True)
class Parameter(AbstractFrozenDataclass):
    name: str

    @abstractmethod
    def get(self) -> Any:
        """
        Returns a parameter value or raise _NeedCalculation if parameter is another piece.
        """


@dataclass(frozen=True, slots=True)
class PieceParameter(Parameter):
    piece_name: str
    type: Type[Any]

    def get(self) -> Any:
        raise _NeedCalculation(self.piece_name, self.type)


@dataclass(frozen=True, slots=True)
class DefaultParameter(Parameter):
    value: Any

    def get(self):
        return self.value


@dataclass(frozen=True, slots=True)
class DefaultFactoryParameter(Parameter):
    factory: Callable[[], Any]

    def get(self):
        return self.factory()
